from typing import Callable
from dataclasses import dataclass
import os
import subprocess
import math
import time
import matplotlib.pyplot as plt
import numpy as np

import gmsh
from lib import sif

def inverse_radius(study):
    aspect = study.aspect

    v = study.var_val
    #lost_area = (4-math.pi) * r*r
    print('meshing aspect = ', aspect)

    # geometry modeling using gmsh
    fact = (math.pi) * v * v
    b = math.sqrt(study.area /(study.aspect - fact)) * 1000
    r = b * v
    a = b * aspect
    #print(a, b, r, v)
    #print(a*b-math.pi*r*r, b/r, v, a/b)
    #exit(0)
    #a = math.sqrt((study.area+lost_area)  * aspect) * 1000
    #b = math.sqrt((study.area+lost_area) / aspect) * 1000

    gmsh.initialize()
    gmsh.option.setNumber("General.Terminal", 1)
    gmsh.model.add("panel")
    geo = gmsh.model.geo

    sz = 5 # 5mm mesh
    # corners - centres of the arcs
    pp1 = geo.add_point(0, 0, 0, sz)
    pp2 = geo.add_point(0, a, 0, sz)
    pp3 = geo.add_point(b, a, 0, sz)
    pp4 = geo.add_point(b, 0, 0, sz)

    p1 = geo.add_point(0, r, 0, sz)
    p2 = geo.add_point(0, a-r, 0, sz)
    p3 = geo.add_point(r, a, 0, sz)
    p4 = geo.addPoint(b-r, a, 0, sz)
    p5 = geo.add_point(b, a-r, 0, sz)
    p6 = geo.add_point(b, r, 0, sz)
    p7 = geo.add_point(b-r, 0, 0, sz)
    p8 = geo.addPoint(r, 0, 0, sz)

    l1 = geo.add_line(p1, p2)
    c1 = geo.add_circle_arc(p2, pp2, p3)
    l2 = geo.add_line(p3, p4)
    c2 = geo.add_circle_arc(p4, pp3, p5)
    l3 = geo.add_line(p5, p6)
    c3 = geo.add_circle_arc(p6, pp4, p7)
    l4 = geo.add_line(p7, p8)
    c4 = geo.add_circle_arc(p8, pp1, p1)


    lp = geo.add_curve_loop([l1, c1, l2, c2, l3, c3, l4, c4])
    surf = geo.add_plane_surface([lp])

    geo.synchronize()
    # generate 2d mesh
    gmsh.model.mesh.generate(2)

    #gmsh.fltk.run()

    gmsh.write("simdata/case.msh")
    gmsh.finalize()


def corner_radius(study):
    aspect = study.aspect

    v = study.var_val
    #lost_area = (4-math.pi) * r*r
    print('meshing aspect = ', aspect)

    # geometry modeling using gmsh
    fact = (4 - math.pi) * v * v
    b = math.sqrt(study.area /(study.aspect - fact)) * 1000
    r = b * v
    a = b * aspect
    #print(a, b, r, v)
    #print(a*b-math.pi*r*r, b/r, v, a/b)
    #exit(0)
    #a = math.sqrt((study.area+lost_area)  * aspect) * 1000
    #b = math.sqrt((study.area+lost_area) / aspect) * 1000

    gmsh.initialize()
    gmsh.option.setNumber("General.Terminal", 1)
    gmsh.model.add("panel")
    geo = gmsh.model.geo

    sz = 5 # 5mm mesh
    # corners - centres of the arcs
    pp1 = geo.add_point(r, r, 0, sz)
    pp2 = geo.add_point(r, a-r, 0, sz)
    pp3 = geo.add_point(b-r, a-r, 0, sz)
    pp4 = geo.add_point(b-r, r, 0, sz)

    p1 = geo.add_point(0, r, 0, sz)
    p2 = geo.add_point(0, a-r, 0, sz)
    p3 = geo.add_point(r, a, 0, sz)
    p4 = geo.addPoint(b-r, a, 0, sz)
    p5 = geo.add_point(b, a-r, 0, sz)
    p6 = geo.add_point(b, r, 0, sz)
    p7 = geo.add_point(b-r, 0, 0, sz)
    p8 = geo.addPoint(r, 0, 0, sz)

    l1 = geo.add_line(p1, p2)
    c1 = geo.add_circle_arc(p2, pp2, p3)
    l2 = geo.add_line(p3, p4)
    c2 = geo.add_circle_arc(p4, pp3, p5)
    l3 = geo.add_line(p5, p6)
    c3 = geo.add_circle_arc(p6, pp4, p7)
    l4 = geo.add_line(p7, p8)
    c4 = geo.add_circle_arc(p8, pp1, p1)


    lp = geo.add_curve_loop([l1, c1, l2, c2, l3, c3, l4, c4])
    surf = geo.add_plane_surface([lp])

    geo.synchronize()
    # generate 2d mesh
    gmsh.model.mesh.generate(2)

    gmsh.fltk.run()

    gmsh.write("simdata/case.msh")
    gmsh.finalize()

def aspect(study):
    aspect = study.var_val
    print('meshing aspect = ', aspect)
    # geometry modeling using gmsh
    
    a = math.sqrt(study.area * aspect) * 1000
    b = math.sqrt(study.area / aspect) * 1000

    gmsh.initialize()
    gmsh.option.setNumber("General.Terminal", 1)
    gmsh.model.add("panel")
    geo = gmsh.model.geo

    sz = 5 # 5mm mesh
    p1 = geo.add_point(0, 0, 0, sz)
    p2 = geo.add_point(0, a, 0, sz)
    p3 = geo.add_point(b, a, 0, sz)
    p4 = geo.addPoint(b, 0, 0, sz)

    l1 = geo.add_line(p1, p2)
    l2 = geo.add_line(p2, p3)
    l3 = geo.add_line(p3, p4)
    l4 = geo.add_line(p4, p1)

    lp = geo.add_curve_loop([l1, l2, l3, l4])
    surf = geo.add_plane_surface([lp])

    geo.synchronize()
    # generate 2d mesh
    gmsh.model.mesh.generate(2)

    #gmsh.fltk.run()

    gmsh.write("simdata/case.msh")
    gmsh.finalize()


def side_slant(study):
    slant = study.var_val
    print('meshing slant= ', slant)

    a = math.sqrt(study.area * study.aspect) * 1000
    b = math.sqrt(study.area / study.aspect) * 1000

    # slant is fraction of side b
    d = slant * b

    gmsh.initialize()
    gmsh.option.setNumber("General.Terminal", 1)
    gmsh.model.add("panel")
    geo = gmsh.model.geo

    sz = 5 # 5mm mesh
    p1 = geo.add_point(0, 0, 0, sz)
    p2 = geo.add_point(0, a, 0, sz)
    p3 = geo.add_point(b-d, a, 0, sz)
    p4 = geo.addPoint(b+d, 0, 0, sz)

    l1 = geo.add_line(p1, p2)
    l2 = geo.add_line(p2, p3)
    l3 = geo.add_line(p3, p4)
    l4 = geo.add_line(p4, p1)

    lp = geo.add_curve_loop([l1, l2, l3, l4])
    surf = geo.add_plane_surface([lp])

    geo.synchronize()
    # generate 2d mesh
    gmsh.model.mesh.generate(2)

    # Uncomment to see the meshed shape in gmsh GUI
    gmsh.fltk.run()

    gmsh.write("simdata/case.msh")
    gmsh.finalize()

@dataclass
class Study:
    area:float = 0.5    # shape area (constant through study)
    aspect:float = 2.0  # shape aspect (constant through study)
    f_max:float = 1000
    shape:Callable = None
    n_vals:int = 101
    var_name:str = 'Side slant'
    var_min:float = 0.0
    var_max:float = 0.9
    var_val:float = 0.0


    def init_study(self):
        sim_dir = "simdata"

        if not os.path.exists(sim_dir):
            os.mkdir(sim_dir)
        
        # write sif file
        with open(os.path.join('simdata', 'case.sif'), "w") as f:
            f.write(sif.sif_template)


    def trim_tail(self, list, max_f):
        '''
        Eliminate data too large for the plot, check that all frequencies are present
        '''

        if list[-1] < max_f:
            raise('Not enough eigenfrequencies')

        j = len(list)-1
        while j > 0 and list[j] > max_f:
            j -= 1

        return j



    def run_once(self):
        '''
        Build the mesh, convert to Elmer format, and run Elmer Solver 
        for one value of the varying value
        '''

        # Mesh the shape
        self.shape(self)
        exceptcount = 0
        while(1):
            try:
                #env = {'PATH' : ''}
                args = ['C:/Program Files/Elmer 9.0-Release/bin/ElmerGrid', '14', '2', 'case.msh', '-autoclean', '-out', 'mesh']

                # generate the Elmer mesh
                #"simdata/logs/elmergrid.log"
                with open('NUL', "w") as f:
                    subprocess.run(args, cwd="simdata", stdout=f, stderr=f, check=True)

                # Run the Elmer solver
                args = ['C:/Program Files/Elmer 9.0-Release/bin/ElmerSolver', "case.sif"]
                #"simdata/logs/elmersolver.log"
                with open('NUL', "w") as f:
                    subprocess.run(args, cwd="simdata", stdout=f, stderr=f, check=True)

                break
            except:
                self.total_fails += 1
                exceptcount += 1
                if exceptcount > 3:
                    raise 'Mesh/solve failed 3 times'
        
        # extract data

        with open('results/eigenfreq.dat') as f:
            eftext = f.readline()
            ef_strings = eftext.strip().split('   ')
            
        eigen_freqs = [float(fstr) for fstr in ef_strings]
        return eigen_freqs


    def plot_modes(self):
        
        var_values = np.linspace(self.var_min, self.var_max, self.n_vals)

        fig, ax = plt.subplots()
        plt.ylabel('Frequency')
        plt.xlabel(self.var_name)
        plt.title('')
        curve_data = [[] for x in range(50)]

        for v in var_values:
            print('run var=', v)
            self.var_val = v
            eigen_freqs = self.run_once()
            #end = trim_tail(eigen_freqs, self.f_max)
            i = 0
            for c in range(0, 50):
                curve_data[c].append(eigen_freqs[i])
                i += 1
            
        for c in range(50):
            ax.plot(var_values, curve_data[c])

        plt.show()


    def plot_stats(self):
        fig, ax = plt.subplots()
        self.total_fails = 0
        var_vals = np.linspace(self.var_min, self.var_max, self.n_vals)

        std_devs = []
        means = []
        outliers = []
        skewness = []
        smalls = []

        for v in var_vals:
            self.var_val = v
            freqs = self.run_once()
            #time.sleep(1.0)
            freqs = freqs[:self.trim_tail(freqs, self.f_max)]
            #print (freqs)
            diffs = [(freqs[i+1]-freqs[i]) for i, _ in enumerate(freqs[0:-1]) ]
            #print (len(diffs))
            mean = sum(diffs) / len(diffs)
            means.append(mean)
            
            sd = math.sqrt(sum([(x-mean)**2 for x in diffs]) / len(diffs))
            std_devs.append(sd)

            sk = sum([(x-mean)**3 for x in diffs]) / (len(diffs)-1) / sd**3
            skewness.append(sk)

            smalls.append(len([x for x in diffs if x < 5])/10)
            
            outliers.append(len([d for d in diffs if abs(d-mean) > sd]))

        print('total fails', self.total_fails)
        ax.grid(True, 'major')
        plt.title(f'Mode gap stats, 0-{self.f_max} Hz')
        plt.xlabel(self.var_name)
        plt.ylabel('Mean (orange) SD (blue) skewness (green')
        ax.plot(var_vals, means)
        ax.plot(var_vals, std_devs)
        ax.plot(var_vals, skewness)
        ax.plot(var_vals, smalls)
        
        plt.show()


    def plot_histogram(self):
        fig, ax = plt.subplots()
 
        
        freqs = self.run_once()
        print (freqs)
        diffs = [(freqs[i+1]-freqs[i]) for i, _ in enumerate(freqs[0:-1]) ]

        x, bins = np.histogram(diffs, bins = 20, range = [0, 20])
        print(x)
        print(bins)
        plt.hist(diffs, bins)

        plt.title(f'Mode gap distribution at {self.var_name} = {self.var_val}')
        plt.xlabel('Gap to next mode')
        plt.ylabel('Frequency')
        #ax.plot(aspects, means)
        plt.show()



st = Study(shape=side_slant, var_min=0.3, n_vals=10)
#st = Study(shape=inverse_radius, var_name='Rounded corners', var_min=0.1, var_max=0.40, n_vals=10)


#st.plot_histogram()
st.plot_stats()

#st.plot_modes()

'''
import vtk

# The source file
file_name = "C:\\Users\\paul\\OneDrive\\Documents\\Elmer\\mypyelmer\\results\\case_t0001.vtu"

# Read the source file.
reader = vtk.vtkXMLUnstructuredGridReader()
reader.SetFileName(file_name)
reader.Update()  # Needed because of GetScalarRange
output = reader.GetOutput()
print(output)
em = output.GetPointData().GetArray("deflection EigenMode1")

print (dir(em))
'''